package TagEngine(
    TagEngine(..),
    mkTagEngine) where

import Vector
import Util

#define UIntLog2N(n) (UInt (TLog n))

interface (TagEngine :: # -> *) numTags =
    requestTag :: ActionValue UIntLog2N(numTags)
    setTagToRetire  :: UIntLog2N(numTags) -> Action
    retireTag  :: Action

-- The tag engine returns a tag that is unique for the duration of
-- the lifetime of the tag. Useful when you need to tag transactions
-- on a bus for example.
-- This implementation is stack based.
mkTagEngine :: Module (TagEngine numTags)
mkTagEngine =
    do
        let reifiedNumTags = fromInteger |> valueOf numTags

        freeStackVec :: Vector numTags (Reg UIntLog2N(numTags))
        freeStackVec <- mapM (\i -> mkReg |> fromInteger i) genVector

        inUseVec :: Vector numTags (Reg Bool)
        inUseVec <- replicateM |> mkReg False

        stackPtr :: (Reg (Maybe(UIntLog2N(numTags))))
        stackPtr <- mkReg |> Just |> reifiedNumTags - 1

        tagToRetire :: RWire (UIntLog2N(numTags))
        tagToRetire <- mkRWireSBR

        debugOnce <- mkReg True

        addRules $
            rules
                "display": when (debugOnce == True) ==>
                    do
                        $display "freeStackVec : " (fshow |> readVReg freeStackVec)
                        $display "inUseVec : " (fshow |> readVReg inUseVec)
                        $display "stackPtr : " (fshow stackPtr)
                        debugOnce := False

        counter <- mkReg(0 :: UIntLog2N(numTags))
        return $
          interface TagEngine

            requestTag :: ActionValue UIntLog2N(numTags)
            requestTag =
                do
                    stackPtr :=
                        if sampledStackPtr == 0
                            then Nothing
                            else Just |> sampledStackPtr - 1
                    (select inUseVec sampledStackPtr) := True
                    return |> readReg (select freeStackVec sampledStackPtr)
                when
                    Just sampledStackPtr <- stackPtr

            setTagToRetire  :: UIntLog2N(numTags) -> Action
            setTagToRetire  tag =
                let
                    tagValid = tag < (reifiedNumTags - 1)
                    tagInUse = readReg (select inUseVec tag)
                in
                    do
                        if (tagValid && tagInUse)
                            then do tagToRetire.wset tag
                            else do action {}

            retireTag :: Action
            retireTag =
                do
                    let nextStackPtrUint =
                            case stackPtr of
                                Nothing -> 0
                                Just n -> n + 1

                    (select freeStackVec nextStackPtrUint) := tag
                    stackPtr := Just nextStackPtrUint
                    (select inUseVec tag) := False
                when
                    Valid tag <- tagToRetire.wget
